#include <opencv2/opencv.hpp>

int main()
{
    cv::Mat part = cv::imread("../iCub-head.png",cv::IMREAD_COLOR);
    cv::Mat whole = cv::imread("../iCub-whole-warped.png", cv::IMREAD_COLOR);

    // create ORB feature detector
    cv::Ptr<cv::Feature2D> featureDetector = cv::SIFT::create();
    //cv::Ptr<cv::Feature2D> featureDetector = cv::ORB::create(500);

    // Detect ORB target features and compute descriptors
    std::vector<cv::KeyPoint> keypoints_part;
    cv::Mat descriptors_part;
    featureDetector->detectAndCompute(part, cv::Mat(), keypoints_part, descriptors_part);

    // Detect ORB features and compute descriptors
    std::vector<cv::KeyPoint> keypoints_whole;
    cv::Mat descriptors_whole;
    featureDetector->detectAndCompute(whole, cv::Mat(), keypoints_whole, descriptors_whole);

    // create descriptors matcher
    cv::Ptr<cv::DescriptorMatcher> matcher = cv::DescriptorMatcher::create(cv::DescriptorMatcher::BRUTEFORCE_L1);
    //cv::Ptr<cv::DescriptorMatcher> matcher = cv::DescriptorMatcher::create(cv::DescriptorMatcher::BRUTEFORCE_HAMMING);

    // Match features
    std::vector<cv::DMatch> matches;
    matcher->match(descriptors_part, descriptors_whole, matches, cv::Mat());

    // Sort matches by score
    std::sort(matches.begin(), matches.end());

    // Remove not so good matches
    int numGoodMatches = int(matches.size() * 0.9);
    matches.resize(numGoodMatches);

    // Draw top matches
    cv::Mat imMatches;
    cv::drawMatches(part, keypoints_part, whole, keypoints_whole, matches, imMatches);

    // Extract location of good matches
    std::vector<cv::Point2f> points_part, points_whole;
    for( size_t i = 0; i < matches.size(); i++ ) {
        points_part.push_back(keypoints_part[matches[i].queryIdx].pt);
        points_whole.push_back(keypoints_whole[matches[i].trainIdx].pt);
    }

    // Find homography
    cv::Mat homography = findHomography(points_part, points_whole, cv::RANSAC);

    // Use homography
    cv::Mat imRegistration;
    warpPerspective(part, imRegistration, homography, whole.size());

    // Display result
    cv::Mat disp_part = part.clone();
    for (cv::Point2f &p : points_part) cv::circle(disp_part,p,1,cv::Scalar(0,255,255),cv::FILLED);
    cv::imwrite("keypoints-part.png", disp_part);
    cv::imshow("keypoints-part", disp_part);
    cv::waitKey(0);

    cv::Mat disp_whole = whole.clone();
    for (cv::Point2f &p : points_whole) cv::circle(disp_whole, p, 1, cv::Scalar(0, 255, 255), cv::FILLED);
    cv::imwrite("keypoints-whole.png", disp_whole);
    cv::imshow("keypoints-whole", disp_whole);
    cv::waitKey(0);

    int height_part = part.rows;
    int width_part = part.cols;
    std::vector<cv::Point2f> points {
        cv::Point2f(0,0), 
        cv::Point2f(float(width_part-1),0),
        cv::Point2f(float(width_part-1),float(height_part-1)), 
        cv::Point2f(0,float(height_part-1))
    };
    std::vector<cv::Point2f> transformed_points_f;
    cv::perspectiveTransform(points,transformed_points_f,homography);
    cv::Rect rect = cv::boundingRect(transformed_points_f);
    std::vector<cv::Point2i> transformed_points;
    for (cv::Point2f &point : transformed_points_f)
        transformed_points.push_back(cv::Point2i(int(point.x+width_part),int(point.y)));
    for (int i=0; i<int(transformed_points.size()); i++)
        cv::line(imMatches,transformed_points[i],transformed_points[(i+1)%int(transformed_points.size())],cv::Scalar(0,255,0),2);
    cv::rectangle(imMatches,cv::Point(width_part+rect.x,rect.y),cv::Point(width_part+rect.x+rect.width,rect.y+rect.height),cv::Scalar(0,0,255),2);
    cv::imwrite("matches.png", imMatches);
    cv::imshow("matches",imMatches);
    cv::waitKey(0);

    cv::imwrite("registration.png",imRegistration);
    cv::imshow("registration",imRegistration);
    cv::waitKey(0);

    cv::Mat gray;
    cv::cvtColor(whole,gray,cv::COLOR_BGR2GRAY);
    cv::cvtColor(gray,gray,cv::COLOR_GRAY2BGR);
    cv::Mat mask;
    cv::inRange(imRegistration, cv::Scalar(0,0,0), cv::Scalar(0,0,0), mask);
    gray.setTo(cv::Scalar(0,0,0),~mask);
    cv::Mat imCombined;
    cv::bitwise_or(gray,imRegistration,imCombined);
    cv::imshow("combined",imCombined);
    cv::waitKey(0);

    cv::destroyAllWindows();
    return 0;
}
